@@ -17,9 +17,8 @@ import tools # since T is used as a variable in this module, refrain from "impor
 import tools.torch
 import tools.modules
 
-# from . import validation as V
 from . import utils as U
-from . import eval as E
+from . import evaluation as E
 from . import model as M
 
 # __all__ = [
@@ -31,9 +30,8 @@ log = logging.getLogger(__name__)
 
 # %%
 class Trainer:
-    def __init__(self, network, train_dataset=None, cfg_train={}, criterion=None, network_dir=None, model_dir=None, cfg_val=None, val_dataset=None, val_metrics=None, name='default', verbose=True, amp=False, reproduce=True):
+    def __init__(self, network, train_dataset=None, cfg_train={}, criterion=None, network_dir=None, cfg_val=None, val_dataset=None, val_metrics=None, name='default', verbose=True, amp=False, reproduce=True):
         '''
-        Things Trainer do:
         - train network with early stopping, given hyperparameters
         - save/load network with necessary preprocessing pipeline
 
@@ -49,6 +47,7 @@ class Trainer:
             lr
             batch_size
             weight_decay (optional)
+            cv_step (optional)
         
         :param cfg_val (optional): dict-like object which contains:
             increase_better
@@ -62,8 +61,8 @@ class Trainer:
         Recommended way of making hs.node.Node object is like the following:
 
             kwargs = {'network': Network(), 'train_dataset': train_dataset, 'validation': None, 'cfg_train': cfg.train,
-                    'criterion': criterion, 'network_dir': path['network'], 'model_dir': path['node'], 'verbose': True, 'amp': True}
-            node = hs.node.Node(**kwargs)
+                    'criterion': criterion, 'network_dir': path['network'], 'verbose': True, 'amp': True}
+            node = hs.Trainer(**kwargs)
         '''
         # Store configurations
         self.network = network
@@ -71,7 +70,6 @@ class Trainer:
         self.cfg_train = dcopy(dict(cfg_train))
         self.criterion = criterion
         self.network_dir = network_dir
-        self.model_dir = model_dir
 
         self.cfg_val = dcopy(dict(cfg_val)) if cfg_val is not None else {}
         self.val_dataset = val_dataset
@@ -92,10 +90,6 @@ class Trainer:
             if os.path.exists(self.network_dir): log.warning(f'path: {self.network_dir} exists. Be careful')
             os.makedirs(self.network_dir, exist_ok=True)
 
-        if self.model_dir is not None:
-            if os.path.exists(self.model_dir): log.warning(f'path: {self.model_dir} exists. Be careful')
-            os.makedirs(self.model_dir, exist_ok=True)
-
         self.train_meter = tools.modules.AverageMeter() # Tracks loss per epoch
         self.loss_tracker = tools.modules.ValueTracker() # Tracks loss over all training
         self.set_model()
@@ -119,7 +113,7 @@ class Trainer:
         self.misc = tools.TDict()
         if self.train_dataset is not None:
             if hasattr(self.train_dataset, 'get_f'): # get_f equal to preprocessing pipeline before feeding to network
-                self.print('get_f found in loader.dataset')
+                self.print('get_f found in train_dataset')
                 self.misc.get_f = self.train_dataset.get_f
 
     def validate(self):
@@ -129,7 +123,7 @@ class Trainer:
             if type(self.val_dataset)==dict:
                 scores = {}
                 for name, dataset in self.val_dataset.items():
-                    d_results = E.test_model(self.model, dataset, batch_size=self.cfg_val['batch_size'], amp=self.amp)
+                    d_results = E.forward_model(self.model, dataset, batch_size=self.cfg_val['batch_size'], amp=self.amp)
                     if any([np.isnan(v).any() for v in d_results.values()]):
                         log.warning(f'[Node: {self.name}][validate] NaN found in d_results. Skipping validation...')
                         score = {self.cfg_val['criterion']: self.earlystopper.best_score}
@@ -139,7 +133,7 @@ class Trainer:
                     self.print(f'[dataset_name: {name}] Validation Score: {score}')
                 score = scores[self.cfg_val['target_dataset']]
             else:
-                d_results = E.test_model(self.model, self.val_dataset, batch_size=self.cfg_val['batch_size'], amp=self.amp)
+                d_results = E.forward_model(self.model, self.val_dataset, batch_size=self.cfg_val['batch_size'], amp=self.amp)
                 if any([np.isnan(v).any() for v in d_results.values()]):
                     log.warning(f'[Node: {self.name}][validate] NaN found in d_results. Skipping validation...')
                     score = {self.cfg_val['criterion']: self.earlystopper.best_score}
@@ -171,10 +165,41 @@ class Trainer:
         else:
             log.warning('No train_dataset found. Skipping generate_loader()')
 
-    def train(self, epoch=None, new_op=True, no_val=False, step=None, reset_loss_tracker=False):
-        '''
+    def train(self, epoch=None, new_op=True, no_val=False, step=None, reset_loss_tracker=False, device=None):
+        """
         Trains the network with the specified duration.
-        '''
+
+        Parameters
+        ----------
+        epoch : int, default=None
+            Number of epochs to train the network.
+            If None, the value is taken from self.cfg_train['epoch']
+            Must be mutually exclusive with "step" argument.
+
+        new_op : bool, default=True
+            Whether to create a new optimizer.
+            If False, the optimizer is not initialized, keeping the previous state.
+
+        no_val : bool, default=False
+            Whether to skip validation during training.
+            If True, the validation is skipped even when val_dataset is given.
+        
+        step : int, default=None
+            Number of steps to train the network.
+            Must be mutually exclusive with "epoch" argument.
+        
+        reset_loss_tracker : bool, default=False
+            Whether to reset the loss_tracker before training.
+
+        device : torch.device, default=None
+            Explicity Device to train the network, recommended to default to None. 
+            If None, the device is taken from the network.
+
+        Returns
+        -------
+        loss_tracker : tools.modules.ValueTracker
+
+        """
         assert epoch is None or step is None, f'only one of epoch or step can be specified. received epoch: {epoch}, step: {step}'
         if step is None:
             horizon = 'epoch'
@@ -190,13 +215,13 @@ class Trainer:
             log.debug(f'[Node: {self.name}] train for {step} steps')
 
         self.network.train()
-        device = tools.torch.get_device(self.network)
-        self.criterion = self.criterion.to(device)
+        self._device = device if device is not None else tools.torch.get_device(self.network)
+        self.criterion = self.criterion.to(self._device)
         if reset_loss_tracker: self.loss_tracker.reset()
         self.generate_loader()
         '''
-        There may be one or more loaders, but self.loader is the standard of synchronization
-        Either return multiple values from dataset, or modify self.forward to use other loaders
+        There may be one or more loaders, but self.loader is the standard of synchronization with batches.
+        Either return multiple values from dataset, or modify self.forward to use other loaders.
         '''
 
         if self.val_dataset is not None:
@@ -209,22 +234,21 @@ class Trainer:
         if new_op or not hasattr(self, 'op'):
             if not hasattr(self, 'op') and not new_op:
                 log.warning("new_op=False when there's no pre-existing optimizer. Ignoring new_op...")
-            # Weight decay optional
-            self.op = optim.Adam(self.network.parameters(), lr=self.cfg_train['lr'], weight_decay=self.cfg_train['weight_decay']) if 'weight_decay' in self.cfg_train \
-                        else optim.Adam(self.network.parameters(), lr=self.cfg_train['lr'])
-            # self.op = optim.Adam(self.network.parameters(), **self.cfg_train)
+            
+            l_kwargs = ['lr', 'weight_decay', 'amsgrad']
+            kwargs = {key: self.cfg_train[key] for key in l_kwargs if key in self.cfg_train}
+            self.op = optim.Adam(self.network.parameters(), **kwargs)
 
         if horizon == 'epoch':
-            self._step_epoch(T=epoch, no_val=no_val, device=device)
+            self._update_epoch(T=epoch, no_val=no_val, device=self._device)
         elif horizon=='step':
-            self._step_step(T=step, no_val=no_val, device=device)
+            self._update_step(T=step, no_val=no_val, device=self._device)
 
         # TODO: Return criterion back to its original device, meaning we have to store its previous device info
         self.criterion = self.criterion.cpu()
         return self.loss_tracker
 
-    def _step_epoch(self, T, no_val=False, device=None):
-        self._device = device if device is not None else tools.torch.get_device(self.network)
+    def _update_epoch(self, T, no_val=False):
 
         _iter = 0
         for epoch in range(1, T+1):
@@ -233,7 +257,8 @@ class Trainer:
             for batch_i, data in enumerate(self.loader, 1):
                 self.iter += 1
                 _iter += 1
-                loss = self._update(data)
+                loss = self.update(data)
+
                 self.print(f'[Node: {self.name}][iter_sum: {self.iter}][Epoch: {epoch}/{T}][Batch: {batch_i}/{len(self.loader)}][Loss: {loss:.7f} (Avg: {self.train_meter.avg:.7f})]')
                 self.loss_tracker.step(self.iter, loss)
 
@@ -244,9 +269,8 @@ class Trainer:
                         self.print('Patience met, stopping training')
                         return None # return None since double break is impossible in python
 
-    def _step_step(self, T, no_val=False, device=None):
-        self._device = device if device is not None else tools.torch.get_device(self.network)
-        if hasattr(self, '_loader_inst'): del self._loader_inst # Load data from the 1st batch
+    def _update_step(self, T, no_val=False):
+        if hasattr(self, '_loader_inst'): del self._loader_inst # Load data from the 1st batch, start fresh
 
         for _iter in range(1, T+1):
             # Get Data
@@ -264,9 +288,9 @@ class Trainer:
                 self.n_batch = 1
 
             self.iter += 1
-            loss = self._update(data)
-            self.print(f'[iter_sum: {self.iter}][Iter: {_iter}/{T}][Batch: {self.n_batch}/{len(self.loader)}][Loss: {loss:.6f} (Avg: {self.train_meter.avg:.6f})]')
+            loss = self.update(data)
 
+            self.print(f'[iter_sum: {self.iter}][Iter: {_iter}/{T}][Batch: {self.n_batch}/{len(self.loader)}][Loss: {loss:.6f} (Avg: {self.train_meter.avg:.6f})]')
             self.loss_tracker.step(self.iter, loss)
 
             # Validation
@@ -277,17 +301,6 @@ class Trainer:
                     self.print('Patience met, stopping training')
                     return None # return None since double break is impossible in python
 
-    def _update(self, data):
-        '''
-        Pseudo function to support amp (automatic mixed precision)
-        '''
-        if self.amp:
-            # Mixed precision for acceleration
-            with torch.autocast(device_type=self._device.type):
-                return self.update(data)
-        else:
-            return self.update(data)
-
     def update(self, data):
         """
         - Perform single update (forward/backward pass + gradient descent step) with the given data.
@@ -306,8 +319,12 @@ class Trainer:
         loss : float
         """
         try:
-            outputs, N = self._forward(data)
-            loss = self._criterion(outputs)
+            # Automatic Mixed Precision (AMP, float32 -> float16) for acceleration
+            if self.amp:
+                with torch.autocast(device_type=self._device.type):
+                    loss, N = self._forward(data)
+            else:
+                loss, N = self._forward(data)
 
             self.op.zero_grad()
             loss.backward()
@@ -321,47 +338,42 @@ class Trainer:
             log.warning(e)
             import pdb; pdb.set_trace()
 
-    def _forward(self, data):
+    def _forward(self, data): # TODO: change to function decorator?
         '''
-        Pseudo function to support amp (automatic mixed precision)
+        Pseudo function to:
+        1. Move data to self._device
+        2. Support passing tuple or dict from a batch from dataloader to forward()
         '''
         datatype = type(data)
         # When data is given as a tuple/list
         if datatype is tuple or datatype is list:
             data = [x.to(self._device) for x in data]
             N = len(data[0]) # number of data in batch
-            outputs = self.forward(*data)
+            loss = self.forward(*data)
 
         # When data is given as a dict
         elif datatype is dict:
             data = {key: value.to(self._device) for key, value in data.items()}
             N = len(next(iter(data.values()))) # number of data in batch
-            outputs = self.forward(**data)
+            loss = self.forward(**data)
 
         else:
             raise Exception(f'return type from dataset must be one of [tuple, list, dict], received: {datatype}')
-        return outputs, N
+        return loss, N
 
     def forward(self, x, y):
         """
-        Forward pass. Receive data and return the loss function.
-        Inherit Node and define new forward() function to build custom forward pass.
+        Forward pass: Receive data and pass values to criterion.
+        Define forward for custom trainers.
 
         May return tuple or dictionary, whichever will be feeded to criterion.
+
+        Defaults to args[0] feeding to self.network (assuming x), and args[1] feeding to self.criterion (assuming y)
         """
-
+        
         y_hat = self.network(x)
+        loss = self.criterion(y_hat, y)
 
-        return y_hat, y
-
-    def _criterion(self, outputs):
-        outputstype = type(outputs)
-        if outputstype is tuple or outputstype is list:
-            loss = self.criterion(*outputs)
-        elif outputstype==dict:
-            loss = self.criterion(**outputs)
-        else:
-            raise Exception(f'return type from forward must be one of [tuple, list, dict], received: {type(data_pred)}')
         return loss
 
     def epoch_f(self):
@@ -451,14 +463,14 @@ class Trainer:
         state_dict() -> path/state_dict.p
         network.state_dict() -> path/network.pt
         '''
-        path = self.model_dir if path is None else path
+        path = self.network_dir if path is None else path
         self.print(f'[Node: {self.name}]')
 
-        state_dict_path = os.path.join(path, 'node.p')
-        self.print(f'[save] Saving node info to: {state_dict_path}')
+        state_dict_path = os.path.join(path, 'model.p')
         state_dict = self.state_dict()
-        self.print(f'[save] state_dict: {list(state_dict.keys())}')
         tools.save_pickle(state_dict, state_dict_path)
+        self.print(f'[save] Saving Model info to: {state_dict_path}')
+        self.print(f'[save] state_dict: {list(state_dict.keys())}')
 
         network_path = os.path.join(path, 'network.pt')
         if best:
@@ -482,7 +494,7 @@ class Trainer:
         path/node.p -> state_dict
         path/network.pt -> network.state_dict
         '''
-        path = self.model_dir if path is None else path
+        path = self.network_dir if path is None else path
         self.print(f'[Node: {self.name}]')
 
         state_dict_path = os.path.join(path, 'node.p')
